# type: ignore
# ruff: noqa
# flake8: noqa
#
# Copyright (c) 2012-2017 The ANTLR Project. All rights reserved.
# Use of this file is governed by the BSD 3-clause license that
# can be found in the LICENSE.txt file in the project root.
from ..atn.ATNState import StarLoopEntryState

from ..atn.ATNConfigSet import ATNConfigSet
from ..atn.ATNState import DecisionState
from .DFAState import DFAState
from ..error.Errors import IllegalStateException


class DFA(object):
    __slots__ = ('atnStartState', 'decision', '_states', 's0', 'precedenceDfa')

    def __init__(self, atnStartState:DecisionState, decision:int=0):
        # From which ATN state did we create this DFA?
        self.atnStartState = atnStartState
        self.decision = decision
        # A set of all DFA states. Use {@link Map} so we can get old state back
        #  ({@link Set} only allows you to see if it's there).
        self._states = dict()
        self.s0 = None
        # {@code true} if this DFA is for a precedence decision; otherwise,
        # {@code false}. This is the backing field for {@link #isPrecedenceDfa},
        # {@link #setPrecedenceDfa}.
        self.precedenceDfa = False

        if isinstance(atnStartState, StarLoopEntryState):
            if atnStartState.isPrecedenceDecision:
                self.precedenceDfa = True
                precedenceState = DFAState(configs=ATNConfigSet())
                precedenceState.edges = []
                precedenceState.isAcceptState = False
                precedenceState.requiresFullContext = False
                self.s0 = precedenceState


    # Get the start state for a specific precedence value.
    #
    # @param precedence The current precedence.
    # @return The start state corresponding to the specified precedence, or
    # {@code null} if no start state exists for the specified precedence.
    #
    # @throws IllegalStateException if this is not a precedence DFA.
    # @see #isPrecedenceDfa()

    def getPrecedenceStartState(self, precedence:int):
        if not self.precedenceDfa:
            raise IllegalStateException("Only precedence DFAs may contain a precedence start state.")

        # s0.edges is never null for a precedence DFA
        if precedence < 0 or precedence >= len(self.s0.edges):
            return None
        return self.s0.edges[precedence]

    # Set the start state for a specific precedence value.
    #
    # @param precedence The current precedence.
    # @param startState The start state corresponding to the specified
    # precedence.
    #
    # @throws IllegalStateException if this is not a precedence DFA.
    # @see #isPrecedenceDfa()
    #
    def setPrecedenceStartState(self, precedence:int, startState:DFAState):
        if not self.precedenceDfa:
            raise IllegalStateException("Only precedence DFAs may contain a precedence start state.")

        if precedence < 0:
            return

        # synchronization on s0 here is ok. when the DFA is turned into a
        # precedence DFA, s0 will be initialized once and not updated again
        # s0.edges is never null for a precedence DFA
        if precedence >= len(self.s0.edges):
            ext = [None] * (precedence + 1 - len(self.s0.edges))
            self.s0.edges.extend(ext)
        self.s0.edges[precedence] = startState
    #
    # Sets whether this is a precedence DFA. If the specified value differs
    # from the current DFA configuration, the following actions are taken;
    # otherwise no changes are made to the current DFA.
    #
    # <ul>
    # <li>The {@link #states} map is cleared</li>
    # <li>If {@code precedenceDfa} is {@code false}, the initial state
    # {@link #s0} is set to {@code null}; otherwise, it is initialized to a new
    # {@link DFAState} with an empty outgoing {@link DFAState#edges} array to
    # store the start states for individual precedence values.</li>
    # <li>The {@link #precedenceDfa} field is updated</li>
    # </ul>
    #
    # @param precedenceDfa {@code true} if this is a precedence DFA; otherwise,
    # {@code false}

    def setPrecedenceDfa(self, precedenceDfa:bool):
        if self.precedenceDfa != precedenceDfa:
            self._states = dict()
            if precedenceDfa:
                precedenceState = DFAState(configs=ATNConfigSet())
                precedenceState.edges = []
                precedenceState.isAcceptState = False
                precedenceState.requiresFullContext = False
                self.s0 = precedenceState
            else:
                self.s0 = None
            self.precedenceDfa = precedenceDfa

    @property
    def states(self):
        return self._states

    # Return a list of all states in this DFA, ordered by state number.
    def sortedStates(self):
        return sorted(self._states.keys(), key=lambda state: state.stateNumber)

    def __str__(self):
        return self.toString(None)

    def toString(self, literalNames:list=None, symbolicNames:list=None):
        if self.s0 is None:
            return ""
        from .DFASerializer import DFASerializer
        serializer = DFASerializer(self,literalNames,symbolicNames)
        return str(serializer)

    def toLexerString(self):
        if self.s0 is None:
            return ""
        from .DFASerializer import LexerDFASerializer
        serializer = LexerDFASerializer(self)
        return str(serializer)



"""
https://datatracker.ietf.org/doc/html/rfc7578
"""
import dataclasses as dc
import io
import typing as ta
import uuid

from .. import cached


MultipartData: ta.TypeAlias = ta.Any  # bytes | file


@dc.dataclass(frozen=True)
class MultipartField:
    data: MultipartData
    name: bytes | None = None
    file_name: bytes | None = None
    headers: ta.Sequence[tuple[bytes, bytes]] | None = None


class MultipartEncoder:
    def __init__(
            self,
            fields: ta.Sequence[MultipartField],
            *,
            boundary: bytes | None = None,
    ) -> None:
        super().__init__()
        self._fields = fields
        self._boundary = boundary or (b'----WebKitFormBoundary-' + uuid.uuid4().hex.encode('ascii'))

    class _Line(ta.NamedTuple):
        data: MultipartData
        sz: int

    @cached.function
    def _lines(self) -> ta.Sequence[_Line]:
        l: list[MultipartEncoder._Line] = []

        def add(d: MultipartData) -> None:
            if isinstance(d, bytes):
                sz = len(d)
            else:
                raise TypeError(d)
            l.append(MultipartEncoder._Line(d, sz))

        for f in self._fields:
            add(b'--%s' % (self._boundary,))
            ps = [b'form-data']
            if f.name is not None:
                ps.append(b'name="%s"' % (f.name,))
            if f.file_name is not None:
                ps.append(b'filename="%s"' % (f.file_name,))
            add(b'Content-Disposition: ' + b'; '.join(ps))
            for hk, hv in f.headers or ():
                add(b'%s: %s' % (hk, hv))
            add(b'')
            add(f.data)

        add(b'--%s--' % (self._boundary,))

        return l

    @cached.function
    def content_type(self) -> bytes:
        return b'multipart/form-data; boundary=%s' % (self._boundary,)

    @cached.function
    def content_length(self) -> int:
        return sum(l.sz + 2 for l in self._lines())

    @cached.function
    def content(self) -> bytes:
        buf = io.BytesIO()
        for l in self._lines():
            if isinstance(l.data, bytes):
                buf.write(l.data)
            else:
                raise TypeError(l.data)
            buf.write(b'\r\n')
        return buf.getvalue()



# type: ignore
# ruff: noqa
# flake8: noqa
#
# Copyright (c) 2012-2017 The ANTLR Project. All rights reserved.
# Use of this file is governed by the BSD 3-clause license that
# can be found in the LICENSE.txt file in the project root.
#/

# A DFA walker that knows how to dump them to serialized strings.#/
from io import StringIO
from ..dfa.DFA import DFA
from ..Utils import str_list
from .DFAState import DFAState


class DFASerializer(object):
    __slots__ = ('dfa', 'literalNames', 'symbolicNames')

    def __init__(self, dfa:DFA, literalNames:list=None, symbolicNames:list=None):
        self.dfa = dfa
        self.literalNames = literalNames
        self.symbolicNames = symbolicNames

    def __str__(self):
        if self.dfa.s0 is None:
            return None
        with StringIO() as buf:
            for s in self.dfa.sortedStates():
                n = 0
                if s.edges is not None:
                    n = len(s.edges)
                for i in range(0, n):
                    t = s.edges[i]
                    if t is not None and t.stateNumber != 0x7FFFFFFF:
                        buf.write(self.getStateString(s))
                        label = self.getEdgeLabel(i)
                        buf.write("-")
                        buf.write(label)
                        buf.write("->")
                        buf.write(self.getStateString(t))
                        buf.write('\n')
            output = buf.getvalue()
            if len(output)==0:
                return None
            else:
                return output

    def getEdgeLabel(self, i:int):
        if i==0:
            return "EOF"
        if self.literalNames is not None and i<=len(self.literalNames):
            return self.literalNames[i-1]
        elif self.symbolicNames is not None and i<=len(self.symbolicNames):
            return self.symbolicNames[i-1]
        else:
            return str(i-1)

    def getStateString(self, s:DFAState):
        n = s.stateNumber
        baseStateStr = ( ":" if s.isAcceptState else "") + "s" + str(n) + ( "^" if s.requiresFullContext else "")
        if s.isAcceptState:
            if s.predicates is not None:
                return baseStateStr + "=>" + str_list(s.predicates)
            else:
                return baseStateStr + "=>" + str(s.prediction)
        else:
            return baseStateStr

class LexerDFASerializer(DFASerializer):

    def __init__(self, dfa:DFA):
        super().__init__(dfa, None)

    def getEdgeLabel(self, i:int):
        return "'" + chr(i) + "'"



import tomllib

from .codecs import make_object_lazy_loaded_codec
from .codecs import make_str_object_codec


##


def _dumps(obj):
    raise TypeError('Unsupported')


TOML_CODEC = make_str_object_codec('toml', _dumps, tomllib.loads)

# @omlish-manifest
_TOML_LAZY_CODEC = make_object_lazy_loaded_codec(__name__, 'TOML_CODEC', TOML_CODEC)



import io
import typing as ta

from ..render import AbstractJsonRenderer
from ..types import SCALAR_TYPES
from .parse import BeginArray
from .parse import BeginObject
from .parse import EndArray
from .parse import EndObject
from .parse import JsonStreamParserEvent
from .parse import Key


##


class StreamJsonRenderer(AbstractJsonRenderer[ta.Iterable[JsonStreamParserEvent]]):
    def __init__(
            self,
            *,
            delimiter: str = '',
            sort_keys: bool = False,
            **kwargs: ta.Any,
    ) -> None:
        if sort_keys:
            raise TypeError('Not yet implemented')

        self._delimiter = delimiter

        super().__init__(**kwargs)

        self._stack: list[tuple[ta.Literal['OBJECT', 'ARRAY'], int]] = []
        self._need_delimit = False

    def _render_value(
            self,
            o: ta.Any,
            state: AbstractJsonRenderer.State = AbstractJsonRenderer.State.VALUE,
    ) -> ta.Generator[str, None, None]:
        if self._style is not None:
            pre, post = self._style(o, state)
            yield pre
        else:
            post = None

        if isinstance(o, SCALAR_TYPES):
            yield self._format_scalar(o)  # type: ignore

        else:
            raise TypeError(o)

        if post:
            yield post

    def _render(self, e: JsonStreamParserEvent) -> ta.Generator[str, None, None]:
        if self._need_delimit:
            yield self._delimiter
            self._need_delimit = False

        if e != EndArray and self._stack and (tt := self._stack[-1])[0] == 'ARRAY':
            if tt[1]:
                yield self._comma
            yield self._get_indent()

            self._stack[-1] = ('ARRAY', tt[1] + 1)

        #

        if e is None or isinstance(e, (str, int, float, bool)):
            yield from self._render_value(e)
            if not self._stack:
                self._need_delimit = True

        #

        elif e is BeginObject:
            self._stack.append(('OBJECT', 0))
            yield '{'
            self._level += 1

        elif isinstance(e, Key):
            if not self._stack or (tt := self._stack.pop())[0] != 'OBJECT':
                raise Exception

            if tt[1]:
                yield self._comma
            yield self._get_indent()
            yield from self._render_value(e.key, AbstractJsonRenderer.State.KEY)
            yield self._colon

            self._stack.append(('OBJECT', tt[1] + 1))

        elif e is EndObject:
            if not self._stack or (tt := self._stack.pop())[0] != 'OBJECT':
                raise Exception

            self._level -= 1
            if tt[1]:
                yield self._get_indent()
            yield '}'
            if not self._stack:
                self._need_delimit = True

        #

        elif e is BeginArray:
            self._stack.append(('ARRAY', 0))
            yield '['
            self._level += 1

        elif e is EndArray:
            if not self._stack or (tt := self._stack.pop())[0] != 'ARRAY':
                raise Exception

            self._level -= 1
            if tt[1]:
                yield self._get_indent()
            yield ']'
            if not self._stack:
                self._need_delimit = True

        #

        else:
            raise TypeError(e)

    def render(self, events: ta.Iterable[JsonStreamParserEvent]) -> ta.Generator[str, None, None]:
        for e in events:
            yield from self._render(e)

    @classmethod
    def render_str(cls, i: ta.Iterable[JsonStreamParserEvent], /, **kwargs: ta.Any) -> str:
        out = io.StringIO()
        for s in cls(**kwargs).render(i):
            out.write(s)
        return out.getvalue()



# ruff: noqa: UP006 UP007
import dataclasses as dc


def dataclass_cache_hash(
        *,
        cached_hash_attr: str = '__dataclass_hash__',
):
    def inner(cls):
        if not isinstance(cls, type) and dc.is_dataclass(cls):
            raise TypeError(cls)

        if (
                cls.__hash__ is object.__hash__ or
                '__hash__' not in cls.__dict__
        ):
            raise TypeError(cls)

        real_hash = cls.__hash__

        def cached_hash(self) -> int:
            try:
                return object.__getattribute__(self, cached_hash_attr)
            except AttributeError:
                object.__setattr__(self, cached_hash_attr, h := real_hash(self))  # type: ignore[call-arg]
            return h

        cls.__hash__ = cached_hash  # type: ignore[method-assign]

        return cls

    return inner



from .base import (  # noqa
    Bootstrap,
    ContextBootstrap,
    SimpleBootstrap,
)

from .diag import (  # noqa
    CheckBootstrap,
    CprofileBootstrap,
    PycharmBootstrap,
    ThreadDumpBootstrap,
    TimebombBootstrap,
)

from .harness import (  # noqa
    bootstrap,
)

from .sys import (  # noqa
    CwdBootstrap,
    EnvBootstrap,
    FaulthandlerBootstrap,
    FdsBootstrap,
    GcBootstrap,
    GcDebugFlag,
    ImportBootstrap,
    LogBootstrap,
    NiceBootstrap,
    PidfileBootstrap,
    PrctlBootstrap,
    PrintPidBootstrap,
    RlimitBootstrap,
    SetuidBootstrap,
)


##


from ..lang.imports import _register_conditional_import  # noqa

_register_conditional_import('..marshal', '.marshal', __package__)



import codecs
import dataclasses as dc
import functools
import typing as ta

from omlish import check

from .base import Codec
from .base import ComboCodec
from .standard import STANDARD_CODECS


##


TextEncodingErrors: ta.TypeAlias = ta.Literal[
    # Raise UnicodeError (or a subclass), this is the default. Implemented in strict_errors().
    'strict',

    # Ignore the malformed data and continue without further notice. Implemented in ignore_errors().
    'ignore',

    # Replace with a replacement marker. On encoding, use ? (ASCII character). On decoding, use � (U+FFFD, the official
    # REPLACEMENT CHARACTER). Implemented in replace_errors().
    'replace',

    # Replace with backslashed escape sequences. On encoding, use hexadecimal form of Unicode code point with formats
    # \xhh \uxxxx \Uxxxxxxxx. On decoding, use hexadecimal form of byte value with format \xhh. Implemented in
    # backslashreplace_errors().
    'backslashreplace',

    # On decoding, replace byte with individual surrogate code ranging from U+DC80 to U+DCFF. This code will then be
    # turned back into the same byte when the 'surrogateescape' error handler is used when encoding the data. (See PEP
    # 383 for more.)
    'surrogateescape',

    ##
    # The following error handlers are only applicable to encoding (within text encodings):

    # Replace with XML/HTML numeric character reference, which is a decimal form of Unicode code point with format
    # &#num;. Implemented in xmlcharrefreplace_errors().
    'xmlcharrefreplace',

    # Replace with \N{...} escape sequences, what appears in the braces is the Name property from Unicode Character
    # Database. Implemented in namereplace_errors().
    'namereplace',

    ##
    # In addition, the following error handler is specific to the given codecs: utf-8, utf-16, utf-32, utf-16-be,
    # utf-16-le, utf-32-be, utf-32-le

    # Allow encoding and decoding surrogate code point (U+D800 - U+DFFF) as normal code point. Otherwise these codecs
    # treat the presence of surrogate code point in str as an error.
    'surrogatepass',
]


@dc.dataclass(frozen=True, kw_only=True)
class TextEncodingOptions:
    errors: TextEncodingErrors = 'strict'


##


class TextEncodingComboCodec(ComboCodec[str, bytes]):
    def __init__(
            self,
            info: codecs.CodecInfo,
            options: TextEncodingOptions = TextEncodingOptions(),
    ) -> None:
        super().__init__()
        self._info = check.isinstance(info, codecs.CodecInfo)
        self._opts = check.isinstance(options, TextEncodingOptions)

    @classmethod
    def lookup(
            cls,
            name: str,
            options: TextEncodingOptions = TextEncodingOptions(),
    ) -> 'TextEncodingComboCodec':
        return cls(codecs.lookup(name), options)

    def encode(self, i: str) -> bytes:
        o, _ = self._info.encode(i, self._opts.errors)
        return o

    def decode(self, o: bytes) -> str:
        i, _ = self._info.decode(o, self._opts.errors)
        return i

    def encode_incremental(self) -> ta.Generator[bytes | None, str, None]:
        x = self._info.incrementalencoder(self._opts.errors)
        i = yield None
        while True:
            if not i:
                break
            o = x.encode(i)
            i = yield o or None
        o = x.encode(i, final=True)
        yield o

    def decode_incremental(self) -> ta.Generator[str | None, bytes, None]:
        x = self._info.incrementaldecoder(self._opts.errors)
        i = yield None
        while True:
            if not i:
                break
            o = x.decode(i)
            i = yield o or None
        o = x.decode(i, final=True)
        yield o


##


class TextEncodingCodec(Codec):
    pass


def normalize_text_encoding_name(s: str) -> str:
    if ' ' in s:
        raise NameError(s)
    return s.lower().replace('_', '-')


def make_text_encoding_codec(
        name: str,
        aliases: ta.Collection[str] | None = None,
        *,
        append_to: ta.MutableSequence[Codec] | None = None,
) -> TextEncodingCodec:
    codec = TextEncodingCodec(
        name=check.equal(name, normalize_text_encoding_name(name)),
        aliases=check.not_isinstance(aliases, str),

        input=str,
        output=bytes,

        new=functools.partial(TextEncodingComboCodec.lookup, name),
        new_incremental=functools.partial(TextEncodingComboCodec.lookup, name),
    )

    if append_to is not None:
        append_to.append(codec)

    return codec


##


ASCII = make_text_encoding_codec(
    'ascii',
    ['646', 'us-ascii'],
    append_to=STANDARD_CODECS,
)

LATIN1 = make_text_encoding_codec(
    'latin-1',
    ['iso-8859-1', 'iso8859-1', '8859', 'cp819', 'latin', 'latin1', 'l1'],
    append_to=STANDARD_CODECS,
)

UTF32 = make_text_encoding_codec(
    'utf-32',
    ['u32', 'utf32'],
    append_to=STANDARD_CODECS,
)

UTF32BE = make_text_encoding_codec(
    'utf-32-be',
    ['utf-32be'],
    append_to=STANDARD_CODECS,
)

UTF32LE = make_text_encoding_codec(
    'utf-32-le',
    ['utf-32le'],
    append_to=STANDARD_CODECS,
)

UTF16 = make_text_encoding_codec(
    'utf-16',
    ['u16', 'utf16'],
    append_to=STANDARD_CODECS,
)

UTF16BE = make_text_encoding_codec(
    'utf-16-be',
    ['utf-16be'],
    append_to=STANDARD_CODECS,
)

UTF16LE = make_text_encoding_codec(
    'utf-16-le',
    ['utf-16le'],
    append_to=STANDARD_CODECS,
)

UTF7 = make_text_encoding_codec(
    'utf-7',
    ['u7', 'unicode-1-1-utf-7'],
    append_to=STANDARD_CODECS,
)

UTF8 = make_text_encoding_codec(
    'utf-8',
    ['u8', 'utf', 'utf8', 'cp65001'],
    append_to=STANDARD_CODECS,
)

UTF8SIG = make_text_encoding_codec(
    'utf-8-sig',
    append_to=STANDARD_CODECS,
)



# type: ignore
# ruff: noqa
# flake8: noqa
#
# Copyright (c) 2012-2017 The ANTLR Project. All rights reserved.
# Use of this file is governed by the BSD 3-clause license that
# can be found in the LICENSE.txt file in the project root.
#/

# Map a predicate to a predicted alternative.#/
from io import StringIO
from ..atn.ATNConfigSet import ATNConfigSet
from ..atn.SemanticContext import SemanticContext


class PredPrediction(object):
    __slots__ = ('alt', 'pred')

    def __init__(self, pred:SemanticContext, alt:int):
        self.alt = alt
        self.pred = pred

    def __str__(self):
        return "(" + str(self.pred) + ", " + str(self.alt) +  ")"

# A DFA state represents a set of possible ATN configurations.
#  As Aho, Sethi, Ullman p. 117 says "The DFA uses its state
#  to keep track of all possible states the ATN can be in after
#  reading each input symbol.  That is to say, after reading
#  input a1a2..an, the DFA is in a state that represents the
#  subset T of the states of the ATN that are reachable from the
#  ATN's start state along some path labeled a1a2..an."
#  In conventional NFA&rarr;DFA conversion, therefore, the subset T
#  would be a bitset representing the set of states the
#  ATN could be in.  We need to track the alt predicted by each
#  state as well, however.  More importantly, we need to maintain
#  a stack of states, tracking the closure operations as they
#  jump from rule to rule, emulating rule invocations (method calls).
#  I have to add a stack to simulate the proper lookahead sequences for
#  the underlying LL grammar from which the ATN was derived.
#
#  <p>I use a set of ATNConfig objects not simple states.  An ATNConfig
#  is both a state (ala normal conversion) and a RuleContext describing
#  the chain of rules (if any) followed to arrive at that state.</p>
#
#  <p>A DFA state may have multiple references to a particular state,
#  but with different ATN contexts (with same or different alts)
#  meaning that state was reached via a different set of rule invocations.</p>
#/
class DFAState(object):
    __slots__ = (
        'stateNumber', 'configs', 'edges', 'isAcceptState', 'prediction',
        'lexerActionExecutor', 'requiresFullContext', 'predicates'
    )

    def __init__(self, stateNumber:int=-1, configs:ATNConfigSet=ATNConfigSet()):
        self.stateNumber = stateNumber
        self.configs = configs
        # {@code edges[symbol]} points to target of symbol. Shift up by 1 so (-1)
        #  {@link Token#EOF} maps to {@code edges[0]}.
        self.edges = None
        self.isAcceptState = False
        # if accept state, what ttype do we match or alt do we predict?
        #  This is set to {@link ATN#INVALID_ALT_NUMBER} when {@link #predicates}{@code !=null} or
        #  {@link #requiresFullContext}.
        self.prediction = 0
        self.lexerActionExecutor = None
        # Indicates that this state was created during SLL prediction that
        # discovered a conflict between the configurations in the state. Future
        # {@link ParserATNSimulator#execATN} invocations immediately jumped doing
        # full context prediction if this field is true.
        self.requiresFullContext = False
        # During SLL parsing, this is a list of predicates associated with the
        #  ATN configurations of the DFA state. When we have predicates,
        #  {@link #requiresFullContext} is {@code false} since full context prediction evaluates predicates
        #  on-the-fly. If this is not null, then {@link #prediction} is
        #  {@link ATN#INVALID_ALT_NUMBER}.
        #
        #  <p>We only use these for non-{@link #requiresFullContext} but conflicting states. That
        #  means we know from the context (it's $ or we don't dip into outer
        #  context) that it's an ambiguity not a conflict.</p>
        #
        #  <p>This list is computed by {@link ParserATNSimulator#predicateDFAState}.</p>
        self.predicates = None



    # Get the set of all alts mentioned by all ATN configurations in this
    #  DFA state.
    def getAltSet(self):
        if self.configs is not None:
            return set(cfg.alt for cfg in self.configs) or None
        return None

    def __hash__(self):
        return hash(self.configs)

    # Two {@link DFAState} instances are equal if their ATN configuration sets
    # are the same. This method is used to see if a state already exists.
    #
    # <p>Because the number of alternatives and number of ATN configurations are
    # finite, there is a finite number of DFA states that can be processed.
    # This is necessary to show that the algorithm terminates.</p>
    #
    # <p>Cannot test the DFA state numbers here because in
    # {@link ParserATNSimulator#addDFAState} we need to know if any other state
    # exists that has this exact set of ATN configurations. The
    # {@link #stateNumber} is irrelevant.</p>
    def __eq__(self, other):
        # compare set of ATN configurations in this set with other
        if self is other:
            return True
        elif not isinstance(other, DFAState):
            return False
        else:
            return self.configs==other.configs

    def __str__(self):
        with StringIO() as buf:
            buf.write(str(self.stateNumber))
            buf.write(":")
            buf.write(str(self.configs))
            if self.isAcceptState:
                buf.write("=>")
                if self.predicates is not None:
                    buf.write(str(self.predicates))
                else:
                    buf.write(str(self.prediction))
            return buf.getvalue()



import struct


def isclose(a: float, b: float, *, rel_tol: float = 1e-09, abs_tol: float = 0.0) -> float:
    return abs(a - b) <= max(rel_tol * max(abs(a), abs(b)), abs_tol)


def float_to_bytes(f: float) -> bytes:
    return struct.pack('>f', f)


def bytes_to_float(b: bytes) -> float:
    return struct.unpack('>f', b)[0]



import typing as ta

import pytest

from ... import lang  # noqa
from .plugins.managermarks import ManagerMark  # noqa


if ta.TYPE_CHECKING:
    from ...asyncs.asyncio import all as aiu
else:
    aiu = lang.proxy_import('...asyncs.asyncio.all', __package__)


class drain_asyncio(ManagerMark):  # noqa
    def __call__(self, item: pytest.Function) -> ta.Iterator[None]:
        with aiu.draining_asyncio_tasks():
            yield



# ruff: noqa: I001
import argparse

from .cli import (  # noqa
    ArgparseArg as Arg,
    argparse_arg as arg,

    ArgparseCommandFn as CommandFn,
    ArgparseCommand as Command,
    argparse_command as command,

    ArgparseCli as Cli,
)


##


SUPPRESS = argparse.SUPPRESS

OPTIONAL = argparse.OPTIONAL
ZERO_OR_MORE = argparse.ZERO_OR_MORE
ONE_OR_MORE = argparse.ONE_OR_MORE
PARSER = argparse.PARSER
REMAINDER = argparse.REMAINDER

HelpFormatter = argparse.HelpFormatter
RawDescriptionHelpFormatter = argparse.RawDescriptionHelpFormatter
RawTextHelpFormatter = argparse.RawTextHelpFormatter
ArgumentDefaultsHelpFormatter = argparse.ArgumentDefaultsHelpFormatter

MetavarTypeHelpFormatter = argparse.MetavarTypeHelpFormatter

ArgumentError = argparse.ArgumentError
ArgumentTypeError = argparse.ArgumentTypeError

Action = argparse.Action
BooleanOptionalAction = argparse.BooleanOptionalAction
SubParsersAction = argparse._SubParsersAction  # noqa

FileType = argparse.FileType

Namespace = argparse.Namespace

ArgumentParser = argparse.ArgumentParser
